@@ -4,7 +4,7 @@ import numpy as np
 import os
 from numba import njit, prange
 
-from .maxcut_tfim_util import get_cut, maxcut_hamming_cdf, opencl_context
+from .maxcut_tfim_util import get_cut, maxcut_hamming_cdf, opencl_context, sample_mag, bit_pick, init_bit_pick
 
 
 epsilon = opencl_context.epsilon
@@ -12,7 +12,7 @@ dtype = opencl_context.dtype
 
 
 @njit
-def update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node):
+def update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node, repulsion_base):
     # Select node
     used[node] = True
 
@@ -20,12 +20,12 @@ def update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node):
     for nbr in range(n):
         if used[nbr]:
             continue
-        weights[nbr] *= max(epsilon, 1 - G_func(nodes[node], nodes[nbr]) / max_edge)
+        weights[nbr] *= repulsion_base ** (-G_func(nodes[node], nodes[nbr]) / max_edge)
 
 
 # Written by Elara (OpenAI custom GPT) and improved by Dan Strano
 @njit
-def local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, n, m):
+def local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, repulsion_base, n, m):
     """
     Pick m nodes out of n with repulsion bias:
     - High-degree nodes are already less likely
@@ -38,72 +38,21 @@ def local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, n,
     used = np.zeros(n, dtype=np.bool_) # False = available, True = used
 
     # First bit:
-    r = np.random.rand()
-    cum = 0.0
-    node = 0
-    for i in range(n):
-        cum += weights[i]
-        if (tot_init_weight * r) < cum:
-            node = i
-            break
+    node = init_bit_pick(weights, tot_init_weight, n)
 
     if m == 1:
         used[node] = True
         return used
 
-    update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node)
+    update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node, repulsion_base)
 
     for _ in range(1, m - 1):
-        # Count available
-        total_w = 0.0
-        for i in range(n):
-            if used[i]:
-                continue
-            total_w += weights[i]
-
-        # Normalize & sample
-        r = np.random.rand()
-        cum = 0.0
-        node = -1
-        for i in range(n):
-            if used[i]:
-                continue
-            cum += weights[i]
-            if (total_w * r) < cum:
-                node = i
-                break
-
-        if node == -1:
-            node = 0
-            while used[node]:
-                node += 1
+        node = bit_pick(weights, used, n)
 
         # Update answer and weights
-        update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node)
+        update_repulsion_choice(G_func, nodes, max_edge, weights, n, used, node, repulsion_base)
 
-    # Count available
-    total_w = 0.0
-    for i in range(n):
-        if used[i]:
-            continue
-        total_w += weights[i]
-
-    # Normalize & sample
-    r = np.random.rand()
-    cum = 0.0
-    node = -1
-    for i in range(n):
-        if used[i]:
-            continue
-        cum += weights[i]
-        if (total_w * r) < cum:
-            node = i
-            break
-
-    if node == -1:
-        node = 0
-        while used[node]:
-            node += 1
+    node = bit_pick(weights, used, n)
 
     # Select node
     used[node] = True
@@ -119,7 +68,7 @@ def compute_energy(sample, G_func, nodes, n_qubits):
             val = G_func(nodes[u], nodes[v])
             energy += val if sample[u] == sample[v] else -val
 
-    return energy
+    return -energy
 
 
 @njit
@@ -134,7 +83,7 @@ def compute_cut(sample, G_func, nodes, n_qubits):
 
 
 @njit(parallel=True)
-def sample_for_energy(G_func, nodes, max_edge, shots, thresholds, degrees_sum, weights, n):
+def sample_measurement(G_func, nodes, max_edge, shots, thresholds, degrees_sum, weights, n, repulsion_base, is_spin_glass):
     shots = max(1, shots >> 1)
     tot_init_weight = weights.sum()
 
@@ -142,69 +91,31 @@ def sample_for_energy(G_func, nodes, max_edge, shots, thresholds, degrees_sum, w
     energies = np.empty(shots, dtype=dtype)
 
     best_solution = solutions[0]
-    best_energy = float("inf")
+    best_energy = -float("inf")
 
     improved = True
     while improved:
         improved = False
         for s in prange(shots):
             # First dimension: Hamming weight
-            mag_prob = np.random.random()
-            m = 0
-            while thresholds[m] < mag_prob:
-                m += 1
-            m += 1
+            m = sample_mag(thresholds)
 
             # Second dimension: permutation within Hamming weight
-            sample = local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, n, m)
+            sample = local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, repulsion_base, n, m)
             solutions[s] = sample
-            energies[s] = compute_energy(sample, G_func, nodes, n)
+            energies[s] = compute_energy(sample, G_func, nodes, n) if is_spin_glass else compute_cut(sample, G_func, nodes, n)
 
-        best_index = np.argmin(energies)
+        best_index = np.argmax(energies)
         energy = energies[best_index]
-        if energy < best_energy:
+        if energy > best_energy:
             best_energy = energy
             best_solution = solutions[best_index].copy()
             improved = True
 
-    return best_solution, compute_cut(best_solution, G_func, nodes, n)
+    if is_spin_glass:
+        best_energy = compute_cut(best_solution, G_func, nodes, n)
 
-
-@njit(parallel=True)
-def sample_for_cut(G_func, nodes, max_edge, shots, thresholds, degrees_sum, weights, n):
-    shots = max(1, shots >> 1)
-    tot_init_weight = weights.sum()
-
-    solutions = np.empty((shots, n), dtype=np.bool_)
-    cuts = np.empty(shots, dtype=dtype)
-
-    best_solution = solutions[0]
-    best_cut = -float("inf")
-
-    improved = True
-    while improved:
-        improved = False
-        for s in prange(shots):
-            # First dimension: Hamming weight
-            mag_prob = np.random.random()
-            m = 0
-            while thresholds[m] < mag_prob:
-                m += 1
-            m += 1
-
-            # Second dimension: permutation within Hamming weight
-            sample = local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, n, m)
-            solutions[s] = sample
-            cuts[s] = compute_cut(sample, G_func, nodes, n)
-
-        best_index = np.argmax(cuts)
-        cut = cuts[best_index]
-        if cut > best_cut:
-            best_cut = cut
-            best_solution = solutions[best_index].copy()
-            improved = True
-
-    return best_solution, best_cut
+    return best_solution, best_energy
 
 
 @njit(parallel=True)
@@ -221,6 +132,7 @@ def init_J_and_z(G_func, nodes):
             if val != 0.0:
                 degree += 1
                 J += val
+                val = abs(val)
             if val > G_max:
                 G_max = val
         J = -J / degree if degree > 0 else 0
@@ -231,18 +143,15 @@ def init_J_and_z(G_func, nodes):
 
 
 @njit
-def cpu_footer(shots, quality, n_qubits, G_func, nodes, is_spin_glass, anneal_t, anneal_h):
-    J_eff, degrees, G_max = init_J_and_z(G_func, nodes)
+def cpu_footer(shots, quality, n_qubits, G_func, nodes, is_spin_glass, anneal_t, anneal_h, repulsion_base):
+    J_eff, degrees, max_edge = init_J_and_z(G_func, nodes)
     hamming_prob = maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, anneal_t, anneal_h)
-    max_edge = degrees.sum()
+    degrees_sum = degrees.sum()
 
     degrees = None
     J_eff = 1.0 / (1.0 + epsilon - J_eff)
 
-    if is_spin_glass:
-        best_solution, best_value = sample_for_energy(G_func, nodes, G_max, shots, hamming_prob, max_edge, J_eff, n_qubits)
-    else:
-        best_solution, best_value = sample_for_cut(G_func, nodes, G_max, shots, hamming_prob, max_edge, J_eff, n_qubits)
+    best_solution, best_value = sample_measurement(G_func, nodes, max_edge, shots, hamming_prob, degrees_sum, J_eff, n_qubits, repulsion_base, is_spin_glass)
 
     bit_string, l, r = get_cut(best_solution, nodes)
 
@@ -256,7 +165,8 @@ def maxcut_tfim_streaming(
     shots=None,
     is_spin_glass=False,
     anneal_t=None,
-    anneal_h=None
+    anneal_h=None,
+    repulsion_base=None
 ):
     wgs = opencl_context.work_group_size
     n_qubits = len(nodes)
@@ -276,7 +186,11 @@ def maxcut_tfim_streaming(
             return "01", weight, ([nodes[0]], [nodes[1]])
 
     if quality is None:
-        quality = 4
+        quality = 6
+
+    if shots is None:
+        # Number of measurement shots
+        shots = n_qubits << quality
 
     if anneal_t is None:
         anneal_t = 8.0
@@ -284,8 +198,13 @@ def maxcut_tfim_streaming(
     if anneal_h is None:
         anneal_h = 8.0
 
-    if shots is None:
-        # Number of measurement shots
-        shots = n_qubits << quality
+    if repulsion_base is None:
+        repulsion_base = 8.0
 
-    return cpu_footer(shots, quality, n_qubits, G_func, nodes, is_spin_glass, anneal_t, anneal_h)
+    bit_string, best_value, partition = cpu_footer(shots, quality, n_qubits, G_func, nodes, is_spin_glass, anneal_t, anneal_h, repulsion_base)
+
+    if best_value < 0.0:
+        # Best cut is trivial partition, all/empty
+        return '0' * n_qubits, 0.0, (nodes, [])
+
+    return bit_string, best_value, partition
