@@ -6,6 +6,7 @@ import networkx as nx
 import numpy as np
 from numba import njit, prange
 import os
+import random
 from scipy.sparse import lil_matrix, csr_matrix
 
 
@@ -42,22 +43,22 @@ def compute_energy(sample, G_data, G_rows, G_cols):
 
 
 @njit
-def bootstrap_worker(theta, G_data, G_rows, G_cols, indices):
+def bootstrap_worker(theta, G_data, G_rows, G_cols, indices, is_spin_glass):
     local_theta = theta.copy()
     for i in indices:
         local_theta[i] = not local_theta[i]
-    energy = compute_energy(local_theta, G_data, G_rows, G_cols)
+    energy = compute_energy(local_theta, G_data, G_rows, G_cols) if is_spin_glass else -evaluate_cut_edges(local_theta, G_data, G_rows, G_cols)
 
     return energy
 
 
 @njit(parallel=True)
-def bootstrap(best_theta, G_data, G_rows, G_cols, indices_array, k, min_energy, dtype):
+def bootstrap(best_theta, G_data, G_rows, G_cols, indices_array, k, min_energy, dtype, is_spin_glass):
     n = len(indices_array) // k
     energies = np.empty(n, dtype=dtype)
     for i in prange(n):
         j = i * k
-        energies[i] = bootstrap_worker(best_theta, G_data, G_rows, G_cols, indices_array[j : j + k])
+        energies[i] = bootstrap_worker(best_theta, G_data, G_rows, G_cols, indices_array[j : j + k], is_spin_glass)
 
     energy = energies.min()
     if energy < min_energy:
@@ -70,7 +71,7 @@ def bootstrap(best_theta, G_data, G_rows, G_cols, indices_array, k, min_energy,
     return min_energy
 
 
-def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices_array_np, k, min_energy, is_segmented, segment_size):
+def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices_array_np, k, min_energy, is_segmented, segment_size, is_spin_glass):
     ctx = opencl_context.ctx
     queue = opencl_context.queue
     bootstrap_kernel = opencl_context.bootstrap_sparse_segmented_kernel if is_segmented else opencl_context.bootstrap_sparse_kernel
@@ -84,7 +85,7 @@ def run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, indices
     best_theta_np = np.array([(1 if b else 0) for b in best_theta], dtype=np.int8)
 
     # Args: [n, k, combo_count, segment_size]
-    args_np = np.array([n, k, combo_count, segment_size], dtype=np.int32)
+    args_np = np.array([n, k, combo_count, is_spin_glass, segment_size], dtype=np.int32)
 
     # Buffers
     mf = cl.mem_flags
@@ -177,6 +178,16 @@ def to_scipy_sparse_upper_triangular(G, nodes, n_nodes, dtype):
     return lil.tocsr()
 
 
+def log_intermediate(theta, G_data, G_rows, G_cols, nodes, min_energy, is_spin_glass):
+    bitstring, l, r = get_cut_from_bit_array(theta, nodes)
+    if is_spin_glass:
+        cut_value = evaluate_cut_edges(theta, G_data, G_rows, G_cols)
+    else:
+        cut_value = -min_energy
+        min_energy = compute_energy(theta, G_data, G_rows, G_cols)
+    print(bitstring, float(cut_value), (l, r), float(min_energy))
+
+
 def spin_glass_solver_sparse(
     G,
     quality=None,
@@ -185,7 +196,12 @@ def spin_glass_solver_sparse(
     is_combo_maxcut_gpu=True,
     is_spin_glass=True,
     anneal_t=None,
-    anneal_h=None
+    anneal_h=None,
+    repulsion_base=None,
+    min_order=1,
+    max_order=None,
+    is_log=False,
+    reheat_tries=0
 ):
     dtype = opencl_context.dtype
     nodes = None
@@ -228,7 +244,7 @@ def spin_glass_solver_sparse(
     elif isinstance(best_guess, list):
         bitstring = "".join(["1" if b else "0" for b in best_guess])
     else:
-        bitstring, _, _ = maxcut_tfim_sparse(G_m, quality=quality, shots=shots, is_spin_glass=is_spin_glass, anneal_t=anneal_t, anneal_h=anneal_h)
+        bitstring, _, _ = maxcut_tfim_sparse(G_m, quality=quality, shots=shots, is_spin_glass=is_spin_glass, anneal_t=anneal_t, anneal_h=anneal_h, repulsion_base=repulsion_base)
     best_theta = np.array([b == "1" for b in list(bitstring)], dtype=np.bool_)
 
     if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
@@ -258,41 +274,76 @@ def spin_glass_solver_sparse(
             else:
                 G_data_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m.data)
 
-    min_energy = compute_energy(best_theta, G_m.data, G_m.indptr, G_m.indices)
-    improved = True
-    correction_quality = 1
+    if max_order is None:
+        max_order = n_qubits
+
+    min_energy = compute_energy(best_theta, G_m.data, G_m.indptr, G_m.indices) if is_spin_glass else -evaluate_cut_edges(best_theta, G_m.data, G_m.indptr, G_m.indices)
+
+    if is_log:
+        log_intermediate(best_theta, G_m.data, G_m.indptr, G_m.indices, nodes, min_energy, is_spin_glass)
+
     combos_list = []
-    while improved:
-        improved = False
-        k = 1
-        while k <= correction_quality:
-            if n_qubits < k:
-                break
+    reheat_theta = best_theta.copy()
+    reheat_min_energy = min_energy
+    for reheat_round in range(reheat_tries + 1):
+        improved = True
+        correction_quality = min_order
+        while improved:
+            improved = False
+            k = 1
+            while (k <= correction_quality) and (k <= max_order):
+                if n_qubits < k:
+                    break
 
-            combos = []
-            if len(combos_list) < k:
-                combos = np.array(list(
-                    item for sublist in itertools.combinations(range(n_qubits), k) for item in sublist
-                ))
-                combos_list.append(combos)
-            else:
-                combos = combos_list[k - 1]
+                combos = []
+                if len(combos_list) < k:
+                    combos = np.array(list(
+                        item for sublist in itertools.combinations(range(n_qubits), k) for item in sublist
+                    ))
+                    combos_list.append(combos)
+                else:
+                    combos = combos_list[k - 1]
 
-            if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
-                energy = run_bootstrap_opencl(best_theta, G_data_buf, G_rows_buf, G_cols_buf, combos, k, min_energy, is_segmented, segment_size)
-            else:
-                energy = bootstrap(best_theta, G_m.data, G_m.indptr, G_m.indices, combos, k, min_energy, dtype)
+                if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
+                    energy = run_bootstrap_opencl(reheat_theta, G_data_buf, G_rows_buf, G_cols_buf, combos, k, reheat_min_energy, is_segmented, segment_size, is_spin_glass)
+                else:
+                    energy = bootstrap(reheat_theta, G_m.data, G_m.indptr, G_m.indices, combos, k, reheat_min_energy, dtype, is_spin_glass)
 
-            if energy < min_energy:
-                min_energy = energy
-                improved = True
-                if correction_quality < (k + 1):
-                    correction_quality = k + 1
-                break
+                if energy < reheat_min_energy:
+                    reheat_min_energy = energy
+                    improved = True
+                    if correction_quality < (k + 1):
+                        correction_quality = k + 1
 
-            k = k + 1
+                    if is_log:
+                        log_intermediate(reheat_theta, G_m.data, G_m.indptr, G_m.indices, nodes, min_energy, is_spin_glass)
+
+                    break
+
+                k = k + 1
+
+        if min_energy < reheat_min_energy:
+            reheat_theta = best_theta.copy()
+        else:
+            best_theta = reheat_theta.copy()
+            min_energy = reheat_min_energy
+
+        if reheat_round < reheat_tries:
+            num_to_flip = int(np.round(np.log2(n_qubits)))
+            bits_to_flip = random.sample(list(range(n_qubits)), num_to_flip)
+            for bit in bits_to_flip:
+                reheat_theta[bit] = not reheat_theta[bit]
+            reheat_min_energy = compute_energy(reheat_theta, G_m.data, G_m.indptr, G_m.indices) if is_spin_glass else -evaluate_cut_edges(reheat_theta, G_m.data, G_m.indptr, G_m.indices)
+
+    opencl_context.G_data_buf = None
+    opencl_context.G_rows_buf = None
+    opencl_context.G_cols_buf = None
 
     bitstring, l, r = get_cut_from_bit_array(best_theta, nodes)
-    cut_value = evaluate_cut_edges(best_theta, G_m.data, G_m.indptr, G_m.indices)
+    if is_spin_glass:
+        cut_value = evaluate_cut_edges(best_theta, G_m.data, G_m.indptr, G_m.indices)
+    else:
+        cut_value = -min_energy
+        min_energy = compute_energy(best_theta, G_m.data, G_m.indptr, G_m.indices)
 
     return bitstring, float(cut_value), (l, r), float(min_energy)
